@@ -9,7 +9,7 @@
    This file is part of Cachegrind, a Valgrind tool for cache
    profiling programs.
 
-   Copyright (C) 2002-2003 Nicholas Nethercote
+   Copyright (C) 2002-2004 Nicholas Nethercote
       njn25@cam.ac.uk
 
    This program is free software; you can redistribute it and/or
@@ -33,8 +33,6 @@
 #include "vg_skin.h"
 //#include "vg_profile.c"
 
-VG_DETERMINE_INTERFACE_VERSION
-
 /* For cache simulation */
 typedef struct {
     int size;       /* bytes */ 
@@ -42,9 +40,7 @@ typedef struct {
     int line_size;  /* bytes */ 
 } cache_t;
 
-#include "cg_sim_L2.c"
-#include "cg_sim_I1.c"
-#include "cg_sim_D1.c"
+#include "cg_sim.c"
 
 /*------------------------------------------------------------*/
 /*--- Constants                                            ---*/
@@ -73,7 +69,7 @@ typedef
       VgpCacheSimulate,
       VgpCacheResults
    } 
-   VgpSkinCC;
+   VgpToolCC;
 
 /*------------------------------------------------------------*/
 /*--- Output file related stuff                            ---*/
@@ -403,6 +399,63 @@ static UInt hash(Char *s, UInt table_size)
     return hash_value;
 }
 
+/* This is a backup for get_BBCC() when removing BBs from the table.
+ * Necessary because the debug info can change when code is removed.  For
+ * example, when inserting, the info might be "myprint.c:myprint()", but
+ * upon removal, the info might be "myprint.c:???", which causes the
+ * hash-lookup to fail (but it doesn't always happen).  So we do a horrible,
+ * slow search through all the file nodes and function nodes (but we can do
+ * 3rd stage with the fast hash-lookup). */
+static BBCC* get_BBCC_slow_removal(Addr bb_orig_addr)
+{
+   Int        i, j;
+   UInt       BBCC_hash;
+   file_node *curr_file_node;
+   fn_node   *curr_fn_node;
+   BBCC     **prev_BBCC_next_ptr, *curr_BBCC;
+
+   for (i = 0; i < N_FILE_ENTRIES; i++) {
+
+      for (curr_file_node = BBCC_table[i]; 
+           NULL != curr_file_node;
+           curr_file_node = curr_file_node->next) 
+      {
+         for (j = 0; j < N_FN_ENTRIES; j++) {
+
+            for (curr_fn_node = curr_file_node->fns[j];
+                 NULL != curr_fn_node;
+                 curr_fn_node = curr_fn_node->next) 
+            {
+               BBCC_hash = bb_orig_addr % N_BBCC_ENTRIES;
+               prev_BBCC_next_ptr = &(curr_fn_node->BBCCs[BBCC_hash]);
+               curr_BBCC = curr_fn_node->BBCCs[BBCC_hash];
+
+               while (NULL != curr_BBCC) {
+                  if (bb_orig_addr == curr_BBCC->orig_addr) {
+                     // Found it!
+                     sk_assert(curr_BBCC->array_size > 0 
+                            && curr_BBCC->array_size < 1000000);
+                     if (VG_(clo_verbosity) > 2) {
+                         VG_(message)(Vg_DebugMsg, "did slow BB removal");
+                     }
+
+                     // Remove curr_BBCC from chain;  it will be used and
+                     // free'd by the caller.
+                     *prev_BBCC_next_ptr = curr_BBCC->next;
+                     return curr_BBCC;
+                  }
+
+                  prev_BBCC_next_ptr = &(curr_BBCC->next);
+                  curr_BBCC = curr_BBCC->next;
+               }
+            }
+         }
+      }
+   }
+   VG_(printf)("failing BB address: %p\n", bb_orig_addr);
+   VG_(skin_panic)("slow BB removal failed");
+}
+
 /* Do a three step traversal: by filename, then fn_name, then instr_addr.
  * In all cases prepends new nodes to their chain.  Returns a pointer to the
  * cost centre.  Also sets BB_seen_before by reference. 
@@ -453,18 +506,23 @@ static BBCC* get_BBCC(Addr bb_orig_addr, UCodeBlock* cb,
    }
    if (curr_BBCC == NULL) {
 
-      sk_assert(False == remove);
-
-      curr_fn_node->BBCCs[BBCC_hash] = curr_BBCC = 
-         new_BBCC(bb_orig_addr, cb, curr_fn_node->BBCCs[BBCC_hash]);
-      *BB_seen_before = False;
+      if (remove == False) {
+         curr_fn_node->BBCCs[BBCC_hash] = curr_BBCC = 
+            new_BBCC(bb_orig_addr, cb, curr_fn_node->BBCCs[BBCC_hash]);
+         *BB_seen_before = False;
+      } else {
+         // Ok, BB not found when removing:  the debug info must have
+         // changed.  Do a slow removal.
+         curr_BBCC = get_BBCC_slow_removal(bb_orig_addr);
+         *BB_seen_before = True;
+      }
 
    } else {
       sk_assert(bb_orig_addr == curr_BBCC->orig_addr);
       sk_assert(curr_BBCC->array_size > 0 && curr_BBCC->array_size < 1000000);
       if (VG_(clo_verbosity) > 2) {
           VG_(message)(Vg_DebugMsg, 
-            "BB retranslation, retrieving from BBCC table");
+            "BB retranslation/invalidation, retrieving from BBCC table");
       }
       *BB_seen_before = True;
 
@@ -548,7 +606,7 @@ static Int compute_BBCC_array_size(UCodeBlock* cb)
 
          case SSE2a_MemRd:
          case SSE2a1_MemRd:
-            sk_assert(u_in->size == 4 || u_in->size == 16);
+            sk_assert(u_in->size == 4 || u_in->size == 8 || u_in->size == 16 || u_in->size == 512);
             t_read = u_in->val3;
             is_FPU_R = True;
             break;
@@ -560,7 +618,7 @@ static Int compute_BBCC_array_size(UCodeBlock* cb)
             break;
 
          case SSE3a1_MemRd:
-            sk_assert(u_in->size == 16);
+            sk_assert(u_in->size == 8 || u_in->size == 16);
             t_read = u_in->val3;
             is_FPU_R = True;
             break;
@@ -581,7 +639,7 @@ static Int compute_BBCC_array_size(UCodeBlock* cb)
             break;
 
          case SSE2a_MemWr:
-            sk_assert(u_in->size == 4 || u_in->size == 16);
+            sk_assert(u_in->size == 4 || u_in->size == 8 || u_in->size == 16 || u_in->size == 512);
             t_write = u_in->val3;
             is_FPU_W = True;
             break;
@@ -802,11 +860,16 @@ UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
 
          case SSE2a_MemRd:
          case SSE2a1_MemRd:
-            sk_assert(u_in->size == 4 || u_in->size == 16);
+            sk_assert(u_in->size == 4 || u_in->size == 8 || u_in->size == 16 || u_in->size == 512);
             t_read = u_in->val3;
             t_read_addr = newTemp(cb);
             uInstr2(cb, MOV, 4, TempReg, u_in->val3,  TempReg, t_read_addr);
-            data_size = u_in->size;
+            /* 512 B data-sized instructions will be done inaccurately
+             * but they're very rare and this avoids errors from
+             * hitting more than two cache lines in the simulation. */
+            data_size = ( u_in->size <= MIN_LINE_SIZE
+                        ? u_in->size
+                        : MIN_LINE_SIZE);
             VG_(copy_UInstr)(cb, u_in);
             break;
 
@@ -820,7 +883,7 @@ UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
             break;
 
          case SSE3a1_MemRd:
-            sk_assert(u_in->size == 16);
+            sk_assert(u_in->size == 8 || u_in->size == 16);
             t_read = u_in->val3;
             t_read_addr = newTemp(cb);
             uInstr2(cb, MOV, 4, TempReg, u_in->val3,  TempReg, t_read_addr);
@@ -860,14 +923,19 @@ UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
             break;
 
          case SSE2a_MemWr:
-            sk_assert(u_in->size == 4 || u_in->size == 16);
+            sk_assert(u_in->size == 4 || u_in->size == 8 || u_in->size == 16 || u_in->size == 512);
            /* fall through */
          case SSE3a_MemWr:
-            sk_assert(u_in->size == 4 || u_in->size == 8 || u_in->size == 16);
+            sk_assert(u_in->size == 4 || u_in->size == 8 || u_in->size == 16 || u_in->size == 512);
             t_write = u_in->val3;
             t_write_addr = newTemp(cb);
             uInstr2(cb, MOV, 4, TempReg, u_in->val3, TempReg, t_write_addr);
-            data_size = u_in->size;
+            /* 512 B data-sized instructions will be done inaccurately
+             * but they're very rare and this avoids errors from
+             * hitting more than two cache lines in the simulation. */
+            data_size = ( u_in->size <= MIN_LINE_SIZE
+                        ? u_in->size
+                        : MIN_LINE_SIZE);
             VG_(copy_UInstr)(cb, u_in);
             break;
 
@@ -2023,7 +2091,7 @@ void SK_(pre_clo_init)(void)
    VG_(details_version)         (NULL);
    VG_(details_description)     ("an I1/D1/L2 cache profiler");
    VG_(details_copyright_author)(
-      "Copyright (C) 2002-2003, and GNU GPL'd, by Nicholas Nethercote.");
+      "Copyright (C) 2002-2004, and GNU GPL'd, by Nicholas Nethercote.");
    VG_(details_bug_reports_to)  (VG_BUGS_TO);
    VG_(details_avg_translation_sizeB) ( 155 );
 
@@ -2072,6 +2140,8 @@ void SK_(post_clo_init)(void)
    init_BBCC_table();
 }
 
+VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 0)
+
 #if 0
 Bool SK_(cheap_sanity_check)(void) { return True; }
 
