@@ -7,7 +7,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -111,7 +111,7 @@ static void pp_Error ( Error* err, Bool printCount )
    }
 }
 
-/* Figure out if we want to attach for GDB for this error, possibly
+/* Figure out if we want to perform a given action for this error, possibly
    by asking the user. */
 Bool VG_(is_action_requested) ( Char* action, Bool* clo )
 {
@@ -191,7 +191,7 @@ void construct_error ( Error* err, ThreadId tid, ErrorKind ekind, Addr a,
    vg_assert( tid < VG_N_THREADS );
 }
 
-void VG_(gen_suppression)(Error* err)
+static void gen_suppression(Error* err)
 {
    Int         i;
    UChar       buf[M_VG_ERRTXT];
@@ -247,30 +247,19 @@ void VG_(gen_suppression)(Error* err)
 }
 
 static 
-void do_actions_on_error(Error* err, Bool allow_GDB_attach)
+void do_actions_on_error(Error* err, Bool allow_db_attach)
 {
-   /* Perhaps we want a GDB attach at this point? */
-   if (allow_GDB_attach &&
-       VG_(is_action_requested)( "Attach to GDB", & VG_(clo_GDB_attach) )) 
+   /* Perhaps we want a debugger attach at this point? */
+   if (allow_db_attach &&
+       VG_(is_action_requested)( "Attach to debugger", & VG_(clo_db_attach) )) 
    {
-      Addr m_eip, m_esp, m_ebp; 
-      
-      if (VG_(is_running_thread)( err->tid )) {
-         m_eip = VG_(baseBlock)[VGOFF_(m_eip)];
-         m_esp = VG_(baseBlock)[VGOFF_(m_esp)];
-         m_ebp = VG_(baseBlock)[VGOFF_(m_ebp)];
-      } else {
-         ThreadState* tst = & VG_(threads)[ err->tid ];
-         m_eip = tst->m_eip;
-         m_esp = tst->m_esp;
-         m_ebp = tst->m_ebp;
-      }
-      VG_(swizzle_esp_then_start_GDB)( m_eip, m_esp, m_ebp );
+      VG_(printf)("starting debugger\n");
+      VG_(start_debugger)( err->tid );
    }
    /* Or maybe we want to generate the error's suppression? */
    if (VG_(is_action_requested)( "Print suppression",
                                  & VG_(clo_gen_suppressions) )) {
-      VG_(gen_suppression)(err);
+      gen_suppression(err);
    }
 }
 
@@ -423,7 +412,7 @@ void VG_(maybe_record_error) ( ThreadId tid,
       pp_Error(p, False);
       is_first_shown_context = False;
       vg_n_errs_shown++;
-      do_actions_on_error(p, /*allow_GDB_attach*/True);
+      do_actions_on_error(p, /*allow_db_attach*/True);
    } else {
       vg_n_errs_suppressed++;
       p->supp->count++;
@@ -431,7 +420,7 @@ void VG_(maybe_record_error) ( ThreadId tid,
 }
 
 /* Second top-level entry point to the error management subsystem, for
-   errors that the skin want to report immediately, eg. because they're
+   errors that the tool wants to report immediately, eg. because they're
    guaranteed to only happen once.  This avoids all the recording and
    comparing stuff.  But they can be suppressed;  returns True if it is
    suppressed.  Bool `print_error' dictates whether to print the error. 
@@ -439,7 +428,7 @@ void VG_(maybe_record_error) ( ThreadId tid,
 */
 Bool VG_(unique_error) ( ThreadId tid, ErrorKind ekind, Addr a, Char* s,
                          void* extra, ExeContext* where, Bool print_error,
-                         Bool allow_GDB_attach, Bool count_error )
+                         Bool allow_db_attach, Bool count_error )
 {
    Error  err;
 
@@ -464,7 +453,7 @@ Bool VG_(unique_error) ( ThreadId tid, ErrorKind ekind, Addr a, Char* s,
          pp_Error(&err, False);
          is_first_shown_context = False;
       }
-      do_actions_on_error(&err, allow_GDB_attach);
+      do_actions_on_error(&err, allow_db_attach);
 
       return False;
 
@@ -638,9 +627,9 @@ static Bool setLocationTy ( Char** p_caller, SuppLocTy* p_ty )
 }
 
 
-/* Look for "skin" in a string like "skin1,skin2,skin3" */
+/* Look for "tool" in a string like "tool1,tool2,tool3" */
 static __inline__
-Bool skin_name_present(Char *name, Char *names)
+Bool tool_name_present(Char *name, Char *names)
 {
    Bool  found;
    Char *s = NULL;   /* Shut gcc up */
@@ -664,7 +653,7 @@ static void load_one_suppressions_file ( Char* filename )
    Int   fd, i;
    Bool  eof;
    Char  buf[N_BUF+1];
-   Char* skin_names;
+   Char* tool_names;
    Char* supp_name;
 
    fd = VG_(open)( filename, VKI_O_RDONLY, 0 );
@@ -675,7 +664,7 @@ static void load_one_suppressions_file ( Char* filename )
    }
 
    while (True) {
-      /* Assign and initialise the two suppression halves (core and skin) */
+      /* Assign and initialise the two suppression halves (core and tool) */
       Supp* supp;
       supp        = VG_(arena_malloc)(VG_AR_CORE, sizeof(Supp));
       supp->count = 0;
@@ -704,11 +693,11 @@ static void load_one_suppressions_file ( Char* filename )
       }
       buf[i]    = '\0';    /* Replace ':', splitting into two strings */
 
-      skin_names = & buf[0];
+      tool_names = & buf[0];
       supp_name  = & buf[i+1];
 
       /* Is it a core suppression? */
-      if (VG_(needs).core_errors && skin_name_present("core", skin_names))
+      if (VG_(needs).core_errors && tool_name_present("core", tool_names))
       {
          if (VG_STREQ(supp_name, "PThread"))
             supp->skind = PThreadSupp;
@@ -716,9 +705,9 @@ static void load_one_suppressions_file ( Char* filename )
             goto syntax_error;
       }
 
-      /* Is it a skin suppression? */
+      /* Is it a tool suppression? */
       else if (VG_(needs).skin_errors && 
-               skin_name_present(VG_(details).name, skin_names))
+               tool_name_present(VG_(details).name, tool_names))
       {
          if (SK_(recognised_suppression)(supp_name, supp)) 
          {
