@@ -9,7 +9,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -244,26 +244,90 @@ Int VG_(gettid)(void)
    mmap/munmap, exit, fcntl
    ------------------------------------------------------------------ */
 
+static Int munmap_inner(void *start, UInt length)
+{
+   return VG_(do_syscall)(__NR_munmap, (UInt)start, (UInt)length );
+}
+
+static Addr mmap_inner(void *start, UInt length, UInt prot, UInt flags, UInt fd, UInt offset)
+{
+   UInt args[6];
+
+   args[0] = (UInt)start;
+   args[1] = length;
+   args[2] = prot;
+   args[3] = flags & ~(VKI_MAP_NOSYMS|VKI_MAP_CLIENT);
+   args[4] = fd;
+   args[5] = offset;
+
+   return VG_(do_syscall)(__NR_mmap, (UInt)(&(args[0])) );   
+}
+
 /* Returns -1 on failure. */
 void* VG_(mmap)( void* start, UInt length, 
                  UInt prot, UInt flags, UInt fd, UInt offset)
 {
-   Int  res;
-   UInt args[6];
-   args[0] = (UInt)start;
-   args[1] = length;
-   args[2] = prot;
-   args[3] = flags;
-   args[4] = fd;
-   args[5] = offset;
-   res = VG_(do_syscall)(__NR_mmap, (UInt)(&(args[0])) );
+   Addr  res;
+
+   if (!(flags & VKI_MAP_FIXED)) {
+      start = (void *)VG_(find_map_space)((Addr)start, length, !!(flags & VKI_MAP_CLIENT));
+      if (start == 0)
+	 return (void *)-1;
+
+      flags |= VKI_MAP_FIXED;
+   }
+
+   res = mmap_inner(start, length, prot, flags, fd, offset);
+
+   if (!VG_(is_kerror)(res)) {
+      UInt sf_flags = SF_MMAP;
+
+      if (flags & VKI_MAP_FIXED)
+	 sf_flags |= SF_FIXED;
+      if (flags & VKI_MAP_SHARED)
+	 sf_flags |= SF_SHARED;
+      if (!(flags & VKI_MAP_ANONYMOUS))
+	 sf_flags |= SF_FILE;
+      if (!(flags & VKI_MAP_CLIENT))
+	 sf_flags |= SF_VALGRIND;
+      if (flags & VKI_MAP_NOSYMS)
+	 sf_flags |= SF_NOSYMS;
+
+      /* placeholder - caller will update flags etc if they want */
+      VG_(map_fd_segment)(res, length, prot, sf_flags, fd, offset, NULL);
+
+      if (flags & VKI_MAP_CLIENT) {
+	 if (res < VG_(client_base) || res >= VG_(client_end)) {
+	    VG_(munmap)((void *)res, length);
+	    res = -1;
+	 }
+      } else {
+	 if (res < VG_(valgrind_base) || res >= VG_(valgrind_end)) {
+	    VG_(munmap)((void *)res, length);
+	    res = -1;
+	 }
+      }
+   }
+
+   
+
    return VG_(is_kerror)(res) ? ((void*)(-1)) : (void*)res;
 }
 
 /* Returns -1 on failure. */
 Int VG_(munmap)( void* start, Int length )
 {
-   Int res = VG_(do_syscall)(__NR_munmap, (UInt)start, (UInt)length );
+   Int res = munmap_inner(start, length);
+   if (!VG_(is_kerror)(res))
+      VG_(unmap_range)((Addr)start, length);
+   return VG_(is_kerror)(res) ? -1 : 0;
+}
+
+Int VG_(mprotect)( void *start, Int length, UInt prot )
+{
+   Int res = VG_(do_syscall)(__NR_mprotect, (UInt)start, (UInt)length, prot );
+   if (!VG_(is_kerror)(res))
+      VG_(mprotect_range)((Addr)start, length, prot);
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
@@ -320,11 +384,49 @@ Int VG_(nanosleep)( const struct vki_timespec *req,
    return 0;
 }
 
+extern Char _end;
+Char *VG_(curbrk) = NULL;
+extern void *__curbrk;		/* in glibc */
+
 void* VG_(brk) ( void* end_data_segment )
 {
-   Int res;
-   res = VG_(do_syscall)(__NR_brk, (UInt)end_data_segment);
-   return (void*)(  VG_(is_kerror)(res) ? -1 : res  );
+   Addr end;
+   Addr brkpage;
+   Addr endpage;
+
+   if (VG_(curbrk) == NULL) {
+      VG_(curbrk) = &_end;
+      __curbrk = (void *)VG_(curbrk);
+   }
+   
+   end = (Addr)end_data_segment;
+   brkpage = PGROUNDUP(VG_(curbrk));
+   endpage = PGROUNDUP(end);
+
+   if (0 && VG_(curbrk) != __curbrk)
+      VG_(printf)("__curbrk changed unexpectedly: VG_(curbrk)=%p, __curbrk=%p\n",
+		  VG_(curbrk), __curbrk);
+
+   if (0)
+      VG_(printf)("brk(end_data_segment=%p); brkpage=%p endpage=%p end=%p curbrk=%p &_end=%p\n",
+		  end_data_segment, brkpage, endpage, end, VG_(curbrk), &_end);
+
+   if (endpage < (Addr)&_end) {
+      __curbrk = (void *)VG_(curbrk);
+      return (void *)VG_(curbrk);
+   }
+
+   if (brkpage != endpage) {
+      if (brkpage > endpage)
+	 munmap_inner((void *)brkpage, brkpage-endpage);
+      else
+	 mmap_inner((void *)brkpage, endpage-brkpage, 
+		    VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC,
+		    VKI_MAP_FIXED|VKI_MAP_PRIVATE|VKI_MAP_ANONYMOUS, -1, 0);
+   }
+   VG_(curbrk) = (Char *)__curbrk = end_data_segment;
+
+   return end_data_segment;
 }
 
 
@@ -968,17 +1070,24 @@ Char VG_(toupper) ( Char c )
 __inline__ Char* VG_(arena_strdup) ( ArenaId aid, const Char* s )
 {
    Int   i;
-   Int   len = VG_(strlen)(s) + 1;
-   Char* res = VG_(arena_malloc) (aid, len);
+   Int   len;
+   Char* res;
+
+   if (s == NULL)
+      return NULL;
+
+   len = VG_(strlen)(s) + 1;
+   res = VG_(arena_malloc) (aid, len);
+
    for (i = 0; i < len; i++)
       res[i] = s[i];
    return res;
 }
 
-/* Wrapper to avoid exposing skins to ArenaId's */
+/* Wrapper to avoid exposing tools to ArenaId's */
 Char* VG_(strdup) ( const Char* s )
 {
-   return VG_(arena_strdup) ( VG_AR_SKIN, s ); 
+   return VG_(arena_strdup) ( VG_AR_TOOL, s ); 
 }
 
 /* ---------------------------------------------------------------------
@@ -992,7 +1101,7 @@ Char* VG_(strdup) ( const Char* s )
 /* Keep track of recursion depth. */
 static Int recDepth;
 
-static Bool string_match_wrk ( Char* pat, Char* str )
+static Bool string_match_wrk ( const Char* pat, const Char* str )
 {
    vg_assert(recDepth >= 0 && recDepth < 500);
    recDepth++;
@@ -1027,7 +1136,7 @@ static Bool string_match_wrk ( Char* pat, Char* str )
    }
 }
 
-Bool VG_(string_match) ( Char* pat, Char* str )
+Bool VG_(string_match) ( const Char* pat, const Char* str )
 {
    Bool b;
    recDepth = 0;
@@ -1057,7 +1166,7 @@ static inline ExeContext *get_real_execontext(Addr ret)
    Addr stacktop;
 
    asm("movl %%ebp, %0; movl %%esp, %1" : "=r" (ebp), "=r" (esp));
-   stacktop = (Addr)&VG_(stack)[VG_STACK_SIZE_W];
+   stacktop = VG_(valgrind_end);
    if (esp >= (Addr)&VG_(sigstack)[0] && esp < (Addr)&VG_(sigstack)[VG_STACK_SIZE_W])
       stacktop = (Addr)&VG_(sigstack)[VG_STACK_SIZE_W];
       
@@ -1144,13 +1253,15 @@ Int VG_(safe_fd)(Int oldfd)
 {
    Int newfd;
 
-   newfd = VG_(fcntl)(oldfd, VKI_F_DUPFD, VG_MAX_FD+1);
+   vg_assert(VG_(max_fd) != -1);
+
+   newfd = VG_(fcntl)(oldfd, VKI_F_DUPFD, VG_(max_fd)+1);
    if (newfd != -1)
       VG_(close)(oldfd);
 
    VG_(fcntl)(newfd, VKI_F_SETFD, VKI_FD_CLOEXEC);
 
-   vg_assert(newfd > VG_MAX_FD);
+   vg_assert(newfd > VG_(max_fd));
    return newfd;
 }
 
@@ -1282,14 +1393,95 @@ Bool VG_(getcwd_alloc) ( Char** out )
    Misc functions looking for a proper home.
    ------------------------------------------------------------------ */
 
-/* We do getenv without libc's help by snooping around in
-   VG_(client_envp) as determined at startup time. */
-Char* VG_(getenv) ( Char* varname )
+/* clone the environment */
+Char **VG_(env_clone) ( Char **oldenv )
+{
+   Char **oldenvp;
+   Char **newenvp;
+   Char **newenv;
+   Int  envlen;
+
+   for (oldenvp = oldenv; oldenvp && *oldenvp; oldenvp++);
+
+   envlen = oldenvp - oldenv + 1;
+   
+   newenv = VG_(arena_malloc)(VG_AR_CORE, envlen * sizeof(Char **));
+
+   oldenvp = oldenv;
+   newenvp = newenv;
+   
+   while (oldenvp && *oldenvp) {
+      *newenvp++ = *oldenvp++;
+   }
+   
+   *newenvp = *oldenvp;
+
+   return newenv;
+}
+
+void  VG_(env_unsetenv) ( Char **env, const Char *varname )
+{
+   Char **from;
+   Char **to = NULL;
+   Int len = VG_(strlen)(varname);
+
+   for(from = to = env; from && *from; from++) {
+      if (!(VG_(strncmp)(varname, *from, len) == 0 && (*from)[len] == '=')) {
+	 *to = *from;
+	 to++;
+      }
+   }
+   *to = *from;
+}
+
+/* set the environment; returns the old env if a new one was allocated */
+Char **VG_(env_setenv) ( Char ***envp, const Char* varname, const Char *val )
+{
+   Char **env = (*envp);
+   Char **cpp;
+   Int len = VG_(strlen)(varname);
+   Char *valstr = VG_(arena_malloc)(VG_AR_CORE, len + VG_(strlen)(val) + 2);
+   Char **oldenv = NULL;
+
+   VG_(sprintf)(valstr, "%s=%s", varname, val);
+
+   for(cpp = env; cpp && *cpp; cpp++) {
+      if (VG_(strncmp)(varname, *cpp, len) == 0 && (*cpp)[len] == '=') {
+	 *cpp = valstr;
+	 return oldenv;
+      }
+   }
+
+   if (env == NULL) {
+      env = VG_(arena_malloc)(VG_AR_CORE, sizeof(Char **) * 2);
+      env[0] = valstr;
+      env[1] = NULL;
+
+      *envp = env;
+
+   }  else {
+      Int envlen = (cpp-env) + 2;
+      Char **newenv = VG_(arena_malloc)(VG_AR_CORE, envlen * sizeof(Char **));
+
+      for(cpp = newenv; *env; )
+	 *cpp++ = *env++;
+      *cpp++ = valstr;
+      *cpp++ = NULL;
+
+      oldenv = *envp;
+
+      *envp = newenv;
+   }
+
+   return oldenv;
+}
+
+Char* VG_(env_getenv) ( Char **env, Char* varname )
 {
    Int i, n;
    n = VG_(strlen)(varname);
-   for (i = 0; VG_(client_envp)[i] != NULL; i++) {
-      Char* s = VG_(client_envp)[i];
+   for (i = 0; env[i] != NULL; i++) {
+      Char* s = env[i];
       if (VG_(strncmp)(varname, s, n) == 0 && s[n] == '=') {
          return & s[n+1];
       }
@@ -1297,6 +1489,12 @@ Char* VG_(getenv) ( Char* varname )
    return NULL;
 }
 
+/* We do getenv without libc's help by snooping around in
+   VG_(client_envp) as determined at startup time. */
+Char *VG_(getenv)(Char *varname)
+{
+   return VG_(env_getenv)(VG_(client_envp), varname);
+}
 
 /* Support for getrlimit. */
 Int VG_(getrlimit) (Int resource, struct vki_rlimit *rlim)
@@ -1309,6 +1507,17 @@ Int VG_(getrlimit) (Int resource, struct vki_rlimit *rlim)
 }
 
 
+/* Support for setrlimit. */
+Int VG_(setrlimit) (Int resource, struct vki_rlimit *rlim)
+{
+   Int res;
+   /* res = setrlimit( resource, rlim ); */
+   res = VG_(do_syscall)(__NR_setrlimit, (UInt)resource, (UInt)rlim);
+   if(VG_(is_kerror)(res)) res = -1;
+   return res;
+}
+
+
 /* Support for getdents. */
 Int VG_(getdents) (UInt fd, struct vki_dirent *dirp, UInt count)
 {
@@ -1363,7 +1572,6 @@ Int VG_(setpgid) ( Int pid, Int pgrp )
 Int VG_(system) ( Char* cmd )
 {
    Int pid, res;
-   void* environ[1] = { NULL };
    if (cmd == NULL)
       return 1;
    pid = VG_(do_syscall)(__NR_fork);
@@ -1371,152 +1579,88 @@ Int VG_(system) ( Char* cmd )
       return -1;
    if (pid == 0) {
       /* child */
+      static Char** envp = NULL;
       Char* argv[4];
+
+      /* restore the DATA rlimit for the child */
+      VG_(setrlimit)(VKI_RLIMIT_DATA, &VG_(client_rlimit_data));
+
+      if (envp == NULL) {
+         Int i;
+         Char* ld_preload_str = NULL;
+         Char* ld_library_path_str = NULL;
+         Char* buf;
+
+         envp = VG_(env_clone)(VG_(client_envp));
+         
+         for (i = 0; envp[i] != NULL; i++) {
+            if (VG_(strncmp)(envp[i], "LD_PRELOAD=", 11) == 0)
+               ld_preload_str = &envp[i][11];
+            if (VG_(strncmp)(envp[i], "LD_LIBRARY_PATH=", 16) == 0)
+               ld_library_path_str = &envp[i][16];
+         }
+
+         buf = VG_(arena_malloc)(VG_AR_CORE, VG_(strlen)(VG_(libdir)) + 20);
+
+         VG_(sprintf)(buf, "%s*/vg_inject.so", VG_(libdir));
+         VG_(mash_colon_env)(ld_preload_str, buf);
+
+         VG_(sprintf)(buf, "%s*/vgpreload_*.so", VG_(libdir));
+         VG_(mash_colon_env)(ld_preload_str, buf);
+
+         VG_(sprintf)(buf, "%s*", VG_(libdir));
+         VG_(mash_colon_env)(ld_library_path_str, buf);
+
+         VG_(env_unsetenv)(envp, VALGRINDCLO);
+
+         VG_(arena_free)(VG_AR_CORE, buf);
+      }
+
       argv[0] = "/bin/sh";
       argv[1] = "-c";
       argv[2] = cmd;
       argv[3] = 0;
+
       (void)VG_(do_syscall)(__NR_execve, 
-			    (UInt)"/bin/sh", (UInt)argv, (UInt)&environ);
+                            (UInt)"/bin/sh", (UInt)argv, (UInt)envp);
+
       /* If we're still alive here, execve failed. */
-      return -1;
+      VG_(exit)(1);
    } else {
       /* parent */
-      res = VG_(do_syscall)(__NR_waitpid, pid, (UInt)NULL, 0);
+      res = VG_(waitpid)(pid, NULL, 0);
       if (VG_(is_kerror)(res)) {
          return -1;
       } else {
-	return 0;
+	 return 0;
       }
    }
 }
 
 
 /* ---------------------------------------------------------------------
-   Support for a millisecond-granularity counter using RDTSC.
+   Support for a millisecond-granularity timer.
    ------------------------------------------------------------------ */
 
-static __inline__ ULong do_rdtsc_insn ( void )
-{
-   ULong x;
-   __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
-   return x;
-}
-
-/* 0 = pre-calibration, 1 = calibration, 2 = running */
-static Int   rdtsc_calibration_state     = 0;
-static ULong rdtsc_ticks_per_millisecond = 0; /* invalid value */
-
-static struct vki_timeval rdtsc_cal_start_timeval;
-static struct vki_timeval rdtsc_cal_end_timeval;
-
-static ULong              rdtsc_cal_start_raw;
-static ULong              rdtsc_cal_end_raw;
-
 UInt VG_(read_millisecond_timer) ( void )
 {
-   ULong rdtsc_now;
-   // If called before rdtsc setup completed (eg. from SK_(pre_clo_init)())
-   // just return 0.
-   if (rdtsc_calibration_state < 2) return 0;
-   rdtsc_now = do_rdtsc_insn();
-   vg_assert(rdtsc_now > rdtsc_cal_end_raw);
-   rdtsc_now -= rdtsc_cal_end_raw;
-   rdtsc_now /= rdtsc_ticks_per_millisecond;
-   return (UInt)rdtsc_now;
-}
-
-
-void VG_(start_rdtsc_calibration) ( void )
-{
+   static ULong base = 0;
+   struct vki_timeval tv_now;
+   ULong now;
    Int res;
-   vg_assert(rdtsc_calibration_state == 0);
-   rdtsc_calibration_state = 1;
-   rdtsc_cal_start_raw = do_rdtsc_insn();
-   res = VG_(do_syscall)(__NR_gettimeofday, (UInt)&rdtsc_cal_start_timeval, 
-			 (UInt)NULL);
-   vg_assert(!VG_(is_kerror)(res));
-}
 
-void VG_(end_rdtsc_calibration) ( void )
-{
-   Int   res, loops;
-   ULong cpu_clock_MHZ;
-   ULong cal_clock_ticks;
-   ULong cal_wallclock_microseconds;
-   ULong wallclock_start_microseconds;
-   ULong wallclock_end_microseconds;
-   struct vki_timespec req;
-   struct vki_timespec rem;
+   res = VG_(do_syscall)(__NR_gettimeofday, (UInt)&tv_now, 
+			 (UInt)NULL);
    
-   vg_assert(rdtsc_calibration_state == 1);
-   rdtsc_calibration_state = 2;
+   now = tv_now.tv_sec * 1000000ULL + tv_now.tv_usec;
+   
+   if (base == 0)
+      base = now;
 
-   /* Try and delay for 20 milliseconds, so that we can at least have
-      some minimum level of accuracy. */
-   req.tv_sec = 0;
-   req.tv_nsec = 20 * 1000 * 1000;
-   loops = 0;
-   while (True) {
-      res = VG_(nanosleep)(&req, &rem);
-      vg_assert(res == 0 /*ok*/ || res == 1 /*interrupted*/);
-      if (res == 0)
-         break;
-      if (rem.tv_sec == 0 && rem.tv_nsec == 0) 
-         break;
-      req = rem;
-      loops++;
-      if (loops > 100) 
-         VG_(core_panic)("calibration nanosleep loop failed?!");
-   }
-
-   /* Now read both timers, and do the Math. */
-   rdtsc_cal_end_raw = do_rdtsc_insn();
-   res = VG_(do_syscall)(__NR_gettimeofday, (UInt)&rdtsc_cal_end_timeval, 
-			 (UInt)NULL);
-
-   vg_assert(rdtsc_cal_end_raw > rdtsc_cal_start_raw);
-   cal_clock_ticks = rdtsc_cal_end_raw - rdtsc_cal_start_raw;
-
-   wallclock_start_microseconds
-      = (1000000ULL * (ULong)(rdtsc_cal_start_timeval.tv_sec)) 
-         + (ULong)(rdtsc_cal_start_timeval.tv_usec);
-   wallclock_end_microseconds
-      = (1000000ULL * (ULong)(rdtsc_cal_end_timeval.tv_sec)) 
-         + (ULong)(rdtsc_cal_end_timeval.tv_usec);
-   vg_assert(wallclock_end_microseconds > wallclock_start_microseconds);
-   cal_wallclock_microseconds 
-      = wallclock_end_microseconds - wallclock_start_microseconds;
-
-   /* Since we just nanoslept for 20 ms ... */
-   vg_assert(cal_wallclock_microseconds >= 20000);
-
-   /* Now we know (roughly) that cal_clock_ticks on RDTSC take
-      cal_wallclock_microseconds elapsed time.  Calculate the RDTSC
-      ticks-per-millisecond value. */
-   if (0)
-      VG_(printf)("%lld ticks in %lld microseconds\n", 
-                  cal_clock_ticks,  cal_wallclock_microseconds );
-
-   rdtsc_ticks_per_millisecond   
-      = cal_clock_ticks / (cal_wallclock_microseconds / 1000ULL);
-   cpu_clock_MHZ
-      = (1000ULL * rdtsc_ticks_per_millisecond) / 1000000ULL;
-   if (VG_(clo_verbosity) >= 1)
-      VG_(message)(Vg_UserMsg, "Estimated CPU clock rate is %d MHz",
-                               (UInt)cpu_clock_MHZ);
-   if (cpu_clock_MHZ < 50 || cpu_clock_MHZ > 10000)
-      VG_(core_panic)("end_rdtsc_calibration: "
-                 "estimated CPU MHz outside range 50 .. 10000");
-   /* Paranoia about division by zero later. */
-   vg_assert(rdtsc_ticks_per_millisecond != 0);
-   if (0)
-      VG_(printf)("ticks per millisecond %llu\n", 
-                  rdtsc_ticks_per_millisecond);
+   return (now - base) / 1000;
 }
 
 
-
 /* ---------------------------------------------------------------------
    Primitive support for bagging memory via mmap.
    ------------------------------------------------------------------ */
@@ -1525,10 +1669,16 @@ void* VG_(get_memory_from_mmap) ( Int nBytes, Char* who )
 {
    static UInt tot_alloc = 0;
    void* p;
-   p = VG_(mmap)( 0, nBytes,
-                     VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC, 
-                     VKI_MAP_PRIVATE|VKI_MAP_ANONYMOUS, -1, 0 );
+   Char *b = VG_(brk)(0);
+
+   p = (void *)PGROUNDUP(b);
+   b = VG_(brk)(p + PGROUNDUP(nBytes));
+
+   if (b != (p + PGROUNDUP(nBytes)))
+       p = (void *)-1;
+
    if (p != ((void*)(-1))) {
+      vg_assert(p >= (void *)VG_(valgrind_mmap_end) && p < (void *)VG_(valgrind_end));
       tot_alloc += (UInt)nBytes;
       if (0)
          VG_(printf)(
