@@ -7,7 +7,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -180,7 +180,7 @@ struct ProxyLWP {
    jmp_buf		jumpbuf;
 };
 
-static void sys_wait_results(Bool block, ThreadId tid, enum RequestType reqtype);
+static void sys_wait_results(Bool block, ThreadId tid, enum RequestType reqtype, Bool restart);
 
 struct PX_Request {
    enum RequestType	request;
@@ -196,7 +196,7 @@ struct PX_Reply {
    union {
       Int		syscallno;	/* system call completed */
       vki_ksiginfo_t	siginfo;	/* signal */
-   };
+   } u;
 };
 
 /* results pipe */
@@ -257,7 +257,6 @@ static const Char *px_name(enum RequestType r)
 }
 
 #define PROXYLWP_OFFSET	(VKI_BYTES_PER_PAGE - sizeof(ProxyLWP))
-#define ROUNDDN(p)	((UChar *)((Addr)(p) & ~(VKI_BYTES_PER_PAGE-1)))
 
 /* 
    Allocate a page for the ProxyLWP and its stack.
@@ -271,7 +270,7 @@ static ProxyLWP *LWP_alloc(void)
 {
    UChar *p = VG_(get_memory_from_mmap)(VKI_BYTES_PER_PAGE, "alloc_LWP");
    ProxyLWP *ret;
-   vg_assert(p == ROUNDDN(p)); /* px must be page aligned */
+   vg_assert(p == (UChar *)PGROUNDDN(p)); /* px must be page aligned */
 
    ret = (ProxyLWP *)(p + PROXYLWP_OFFSET);
 
@@ -283,7 +282,7 @@ static ProxyLWP *LWP_alloc(void)
 /* Free a thread structure */
 static void LWP_free(ProxyLWP *px)
 {
-   UChar *p = ROUNDDN(px);
+   UChar *p = (UChar *)PGROUNDDN(px);
    
    vg_assert(px->magic == VG_PROXY_MAGIC);
    px->magic = 0;
@@ -297,7 +296,7 @@ static void LWP_free(ProxyLWP *px)
    end). */
 static inline ProxyLWP *LWP_TSD(void *esp)
 {
-   UChar *p = ROUNDDN(esp);
+   UChar *p = (UChar *)PGROUNDDN(esp);
    ProxyLWP *ret;
 
    ret = (ProxyLWP *)(p + PROXYLWP_OFFSET);
@@ -512,7 +511,7 @@ static Int proxylwp(void *v)
 	    /* First, send the signal info */
 	    sigreply.tid = px->tid;
 	    sigreply.req = PX_Signal;
-	    sigreply.siginfo = px->siginfo;
+	    sigreply.u.siginfo = px->siginfo;
 
 	    if (!send_reply(&sigreply)) {
 	       ret = 44;		/* incomplete or failed write */
@@ -520,11 +519,12 @@ static Int proxylwp(void *v)
 	    }
 	 } else {
 	    /* We got VKI_SIGVGKILL, which means we just skip all the
-	       below and get back to the state machine - probably to
-	       exit. */
+	       below and exit.  (Don't bother dealing with any pending
+	       requests, because we'll probably just get confused.) */
 	    px->state = PXS_WaitReq;
 	    px->siginfo.si_signo = 0;
-	    goto state_machine;
+	    ret = 0;
+	    goto out;
 	 }
 
 	 px->siginfo.si_signo = 0;
@@ -554,7 +554,7 @@ static Int proxylwp(void *v)
 		  XXX how to distunguish between restartable and
 		  non-restartable syscalls?  Does it matter?
 	       */
-	       reply.syscallno = tst->syscallno;
+	       reply.u.syscallno = tst->syscallno;
 
 	       tst->m_eax = -VKI_ERESTARTSYS;
 	       px->state = PXS_IntReply;
@@ -599,7 +599,7 @@ static Int proxylwp(void *v)
 	    /* We were actually running the syscall when interrupted.
 	       reply should already be set up, including return in eax. */
 	    vg_assert(reply.req == PX_RunSyscall);
-	    vg_assert(reply.syscallno == tst->syscallno);
+	    vg_assert(reply.u.syscallno == tst->syscallno);
 	    vg_assert(tst->status == VgTs_WaitSys);
 	    px->state = PXS_IntReply;
 	    break;
@@ -608,7 +608,7 @@ static Int proxylwp(void *v)
 	    /* The syscall is done; we just need to send the results
 	       back. */
 	    vg_assert(reply.req == PX_RunSyscall);
-	    vg_assert(reply.syscallno == tst->syscallno);
+	    vg_assert(reply.u.syscallno == tst->syscallno);
 	    px->state = PXS_IntReply;
 	    break;
 
@@ -629,7 +629,7 @@ static Int proxylwp(void *v)
 	    VG_(need_resched)(px->tid);
       }
 
-     state_machine:
+      /* state_machine: */
       px_printf("proxylwp main: state %s\n", pxs_name(px->state));
 
       switch(px->state) {
@@ -712,7 +712,7 @@ static Int proxylwp(void *v)
 	 case PX_RunSyscall:
 	    /* Run a syscall for our thread; results will be poked
 	       back into tst */
-	    reply.syscallno = tst->syscallno;
+	    reply.u.syscallno = tst->syscallno;
 
 	    vg_assert(px->state == PXS_WaitReq || 
 		      px->state == PXS_SigACK);
@@ -723,7 +723,7 @@ static Int proxylwp(void *v)
 		  on the way to us as we got the signal.  
 	       */
 	       px_printf("RunSyscall in SigACK: rejecting syscall %d with ERESTARTSYS\n",
-			 reply.syscallno);
+			 reply.u.syscallno);
 	       tst->m_eax = -VKI_ERESTARTSYS;
 	    } else {
 	       Int syscallno = tst->syscallno;
@@ -842,9 +842,9 @@ void VG_(proxy_sendsig)(ThreadId tid, Int sig)
 	 it by the normal mechanism.  In this case, just wait for the
 	 syscall to dinish. */
       if (tst->status == VgTs_WaitSys && tst->syscallno == __NR_rt_sigtimedwait)
-	 sys_wait_results(True, tid, PX_RunSyscall);
+	 sys_wait_results(True, tid, PX_RunSyscall, True);
       else
-	 sys_wait_results(True, tid, PX_Signal);
+	 sys_wait_results(True, tid, PX_Signal, True);
    }
 }
 
@@ -867,7 +867,7 @@ void VG_(proxy_abort_syscall)(ThreadId tid)
    if (lwp != 0)
       VG_(ktkill)(lwp, VKI_SIGVGINT);
 
-   sys_wait_results(True, tid, PX_RunSyscall);
+   sys_wait_results(True, tid, PX_RunSyscall, False);
 
    vg_assert(tst->status == VgTs_Runnable);
 }
@@ -1064,7 +1064,7 @@ void VG_(proxy_delete)(ThreadId tid, Bool force)
       proxy_wait, because if we don't read the results pipe, the proxy
       may be blocked writing to it, causing a deadlock with us as we
       wait for it to exit. */
-   sys_wait_results(True, tid, PX_Exiting);
+   sys_wait_results(True, tid, PX_Exiting, True);
    res = proxy_wait(proxy, True, &status);
 
    if ((!res || status != 0) && VG_(clo_verbosity) > 1)
@@ -1092,7 +1092,7 @@ void VG_(proxy_delete)(ThreadId tid, Bool force)
    reply, otherwise it will block forever).  If tid != 0, then it will
    wait for a reply for that particular tid.
  */
-static void sys_wait_results(Bool block, ThreadId tid, enum RequestType reqtype)
+static void sys_wait_results(Bool block, ThreadId tid, enum RequestType reqtype, Bool restart)
 {
    Bool found_reply = (reqtype == PX_BAD);
    struct PX_Reply res;
@@ -1151,21 +1151,21 @@ static void sys_wait_results(Bool block, ThreadId tid, enum RequestType reqtype)
 	       VG_(printf)("tid %d in status %d\n",
 			   tst->tid, tst->status);
 	 
-	    vg_assert(res.syscallno == tst->syscallno);
+	    vg_assert(res.u.syscallno == tst->syscallno);
 	    vg_assert(tst->status == VgTs_WaitSys);
 
-	    VG_(post_syscall)(res.tid);
+	    VG_(post_syscall)(res.tid, restart);
 	    break;
 
 	 case PX_Signal:
 	    if (VG_(clo_trace_signals) || VG_(clo_trace_syscalls))
 	       VG_(message)(Vg_DebugMsg, "sys_wait_results: got PX_Signal for TID %d, signal %d",
-			    res.tid, res.siginfo.si_signo);
+			    res.tid, res.u.siginfo.si_signo);
 
-	    vg_assert(res.siginfo.si_signo != 0);
+	    vg_assert(res.u.siginfo.si_signo != 0);
 	    if (VG_(threads)[res.tid].proxy && 
 		!VG_(threads)[res.tid].proxy->terminating)
-	       VG_(deliver_signal)(res.tid, &res.siginfo, True);
+	       VG_(deliver_signal)(res.tid, &res.u.siginfo, True);
 	    break;
 
 	 case PX_Ping:
@@ -1188,16 +1188,16 @@ static void sys_wait_results(Bool block, ThreadId tid, enum RequestType reqtype)
 /* External version */
 void VG_(proxy_results)(void)
 {
-   sys_wait_results(False, 0, PX_BAD);
+   sys_wait_results(False, 0, PX_BAD, True);
 }
 
-void VG_(proxy_wait_sys)(ThreadId tid)
+void VG_(proxy_wait_sys)(ThreadId tid, Bool restart)
 {
    ThreadState *tst = VG_(get_ThreadState)(tid);
 
    vg_assert(tst->status == VgTs_WaitSys);
 
-   sys_wait_results(True, tid, PX_RunSyscall);
+   sys_wait_results(True, tid, PX_RunSyscall, restart);
 }
 
 /* Tell proxy about it's thread's updated signal mask */
@@ -1227,7 +1227,7 @@ void VG_(proxy_setsigmask)(ThreadId tid)
       with respect to each other (ie, if thread A then thread B
       updates their signal masks, A's update must be done before B's
       is).  */
-   sys_wait_results(True, tid, PX_SetSigmask);
+   sys_wait_results(True, tid, PX_SetSigmask, True);
 }
 
 void VG_(proxy_sigack)(ThreadId tid, const vki_ksigset_t *mask)
@@ -1272,7 +1272,7 @@ void VG_(proxy_waitsig)(void)
    if (VG_(do_signal_routing))
       VG_(route_signals)();
    else
-      sys_wait_results(True, VG_INVALID_THREADID /* any */, PX_Signal);
+      sys_wait_results(True, VG_INVALID_THREADID /* any */, PX_Signal, True);
 }
 
 /* Issue a syscall to the thread's ProxyLWP */
@@ -1365,7 +1365,7 @@ void VG_(proxy_sanity)(void)
 			 tid, px->lwp, ret);
 	    sane = False;
 	 }
-	 sys_wait_results(True, tid, PX_Ping);
+	 sys_wait_results(True, tid, PX_Ping, True);
 	 /* Can't make an assertion here, fortunately; this will
 	    either come back or it won't. */
       }
@@ -1374,6 +1374,15 @@ void VG_(proxy_sanity)(void)
    vg_assert(sane);
 }
 
+/* Get the PID/TID of the ProxyLWP. */
+Int VG_(proxy_id)(ThreadId tid)
+{
+   ThreadState *tst = VG_(get_ThreadState)(tid);
+   ProxyLWP *proxy = tst->proxy;
+
+   return proxy->lwp;
+}
+
 /*--------------------------------------------------------------------*/
 /*--- Proxy LWP machinery.                           vg_proxylwp.c ---*/
 /*--------------------------------------------------------------------*/
