@@ -8,7 +8,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -984,13 +984,13 @@ void vg_push_signal_frame ( ThreadId tid, const vki_ksiginfo_t *siginfo )
             "delivering signal %d (%s) to thread %d: on ALT STACK", 
             sigNo, signame(sigNo), tid );
 
-      /* Signal delivery to skins */
+      /* Signal delivery to tools */
       VG_TRACK( pre_deliver_signal, tid, sigNo, /*alt_stack*/False );
       
    } else {
       esp_top_of_frame = tst->m_esp;
 
-      /* Signal delivery to skins */
+      /* Signal delivery to tools */
       VG_TRACK( pre_deliver_signal, tid, sigNo, /*alt_stack*/True );
    }
 
@@ -1010,7 +1010,7 @@ void vg_push_signal_frame ( ThreadId tid, const vki_ksiginfo_t *siginfo )
    /* retaddr, sigNo, psigInfo, puContext fields are to be written */
    VG_TRACK( pre_mem_write, Vg_CoreSignal, tid, "signal handler frame", 
                             (Addr)frame, offsetof(VgSigFrame, handlerArgs) );
-   frame->retaddr    = (UInt)(&VG_(signalreturn_bogusRA));
+   frame->retaddr    = (UInt)VG_(client_trampoline_code)+VG_(tramp_sigreturn_offset);
    frame->sigNo      = sigNo;
    frame->sigNo_private = sigNo;
    VG_TRACK( post_mem_write, (Addr)frame, offsetof(VgSigFrame, handlerArgs) );
@@ -1170,7 +1170,7 @@ Int vg_pop_signal_frame ( ThreadId tid )
    tst->sig_mask  = frame->mask;
    VG_(proxy_setsigmask)(tid);
 
-   /* Notify skins */
+   /* Notify tools */
    VG_TRACK( post_deliver_signal, tid, sigNo );
 
    return sigNo;
@@ -1323,6 +1323,10 @@ static void vg_default_action(const vki_ksiginfo_t *info, ThreadId tid)
 
    vg_assert(!core || (core && terminate));
 
+   if (VG_(clo_trace_signals))
+      VG_(message)(Vg_DebugMsg, "delivering %d to default handler %s%s",
+		   sigNo, terminate ? "terminate" : "", core ? "+core" : "");
+
    if (terminate) {
       if (VG_(clo_verbosity) != 0 && (core || VG_(clo_verbosity) > 1)) {
 	 VG_(message)(Vg_UserMsg, "");
@@ -1336,8 +1340,8 @@ static void vg_default_action(const vki_ksiginfo_t *info, ThreadId tid)
 	    switch(sigNo) {
 	    case VKI_SIGSEGV:
 	       switch(info->si_code) {
-	       case 1: event = "Address not mapped to object"; break;
-	       case 2: event = "Invalid permissions for mapped object"; break;
+	       case 1: event = "Access not within mapped region"; break;
+	       case 2: event = "Bad permissions for mapped region"; break;
 	       }
 	       break;
 
@@ -1387,9 +1391,8 @@ static void vg_default_action(const vki_ksiginfo_t *info, ThreadId tid)
 	 }
       }
 
-      if (VG_(is_action_requested)( "Attach to GDB", & VG_(clo_GDB_attach) )) {
-         ThreadState* tst = & VG_(threads)[ tid ];      
-	 VG_(swizzle_esp_then_start_GDB)( tst->m_eip, tst->m_esp, tst->m_ebp );
+      if (VG_(is_action_requested)( "Attach to debugger", & VG_(clo_db_attach) )) {
+         VG_(start_debugger)( tid );
       }
 
       if (VG_(fatal_signal_set)) {
@@ -1409,6 +1412,7 @@ void VG_(deliver_signal) ( ThreadId tid, const vki_ksiginfo_t *info, Bool async
    Int			sigNo = info->si_signo;
    vki_ksigset_t	handlermask;
    SCSS_Per_Signal	*handler = &vg_scss.scss_per_sig[sigNo];
+   void			*handler_fn;
    ThreadState		*tst = VG_(get_ThreadState)(tid);
 
    if (VG_(clo_trace_signals))
@@ -1443,32 +1447,27 @@ void VG_(deliver_signal) ( ThreadId tid, const vki_ksiginfo_t *info, Bool async
 	 came.  Either way, it is going to give us some syscall
 	 results right now, so wait for them to appear.  This makes
 	 the thread runnable again, so we're in the right state to run
-	 the handler, and resume the syscall when we're done. */
-      VG_(proxy_wait_sys)(tid);
-
+	 the handler.  We ask post_syscall to restart based on the
+	 client's sigaction flags. */
       if (0)
-	 VG_(printf)("signal %d interrupting syscall %d\n",
-		     sigNo, tst->syscallno);
-
-      if (tst->m_eax == -VKI_ERESTARTSYS) {
-	  if (handler->scss_flags & VKI_SA_RESTART) {
-	     VG_(restart_syscall)(tid);
-	  } else
-	     tst->m_eax = -VKI_EINTR;
-      } else {
-	 /* return value is already in eax - either EINTR or the
-	    normal return value */
-      }
+	 VG_(printf)("signal %d interrupted syscall %d; restart=%d\n",
+		     sigNo, tst->syscallno, !!(handler->scss_flags & VKI_SA_RESTART));
+      VG_(proxy_wait_sys)(tid, !!(handler->scss_flags & VKI_SA_RESTART));
    }
 
-   vg_assert(handler->scss_handler != VKI_SIG_IGN);
+   /* If the client specifies SIG_IGN, treat it as SIG_DFL */
+   handler_fn = handler->scss_handler;
+   if (handler_fn == VKI_SIG_IGN)
+      handler_fn = VKI_SIG_DFL;
+
+   vg_assert(handler_fn != VKI_SIG_IGN);
 
    if (sigNo == VKI_SIGCHLD && (handler->scss_flags & VKI_SA_NOCLDWAIT)) {
       //VG_(printf)("sigNo==SIGCHLD and app asked for NOCLDWAIT\n");
       vg_babyeater(sigNo, NULL, NULL);
    }
 
-   if (handler->scss_handler == VKI_SIG_DFL) {
+   if (handler_fn == VKI_SIG_DFL) {
       handlermask = tst->sig_mask; /* no change to signal mask */
       vg_default_action(info, tid);
    } else {
@@ -1604,8 +1603,7 @@ void vg_sync_signalhandler ( Int sigNo, vki_ksiginfo_t *info, struct vki_ucontex
    */
 
    if (VG_(clo_trace_signals)) {
-      VG_(start_msg)(Vg_DebugMsg);
-      VG_(add_to_msg)("signal %d arrived ... ", sigNo );
+      VG_(message)(Vg_DebugMsg, "signal %d arrived ... si_code=%d", sigNo, info->si_code );
    }
    vg_assert(sigNo >= 1 && sigNo <= VKI_KNSIG);
 
@@ -1631,6 +1629,77 @@ void vg_sync_signalhandler ( Int sigNo, vki_ksiginfo_t *info, struct vki_ucontex
    vg_assert((Char*)(&(VG_(sigstack)[0])) <= (Char*)(&dummy_local));
    vg_assert((Char*)(&dummy_local) < (Char*)(&(VG_(sigstack)[VG_SIGSTACK_SIZE_W])));
 
+   /* Special fault-handling case. We can now get signals which can
+      act upon and immediately restart the faulting instruction.
+    */
+   if (info->si_signo == VKI_SIGSEGV) {
+      ThreadId tid = VG_(get_current_or_recent_tid)();
+      Addr fault = (Addr)info->_sifields._sigfault._addr;
+      Addr esp = VG_(is_running_thread)(tid) ?
+	 VG_(baseBlock)[VGOFF_(m_esp)] : VG_(threads)[tid].m_esp;
+      Segment *seg;
+
+      seg = VG_(find_segment)(fault);
+      if (seg != NULL)
+	 seg = VG_(next_segment)(seg);
+
+      if (VG_(clo_trace_signals)) {
+	 if (seg == NULL)
+	    VG_(message)(Vg_DebugMsg,
+			 "SIGSEGV: si_code=%d faultaddr=%p tid=%d esp=%p seg=NULL shad=%p-%p",
+			 info->si_code, fault, tid, esp,
+			 VG_(shadow_base), VG_(shadow_end));
+	 else
+	    VG_(message)(Vg_DebugMsg,
+			 "SIGSEGV: si_code=%d faultaddr=%p tid=%d esp=%p seg=%p-%p fl=%x shad=%p-%p",
+			 info->si_code, fault, tid, esp, seg->addr, seg->addr+seg->len, seg->flags,
+			 VG_(shadow_base), VG_(shadow_end));
+      }
+
+      if (info->si_code == 1		&&	/* SEGV_MAPERR */
+	  seg != NULL                   &&
+	  fault >= esp			&&
+	  fault < seg->addr		&&
+	  (seg->flags & SF_GROWDOWN)) {
+	 /* If the fault address is above esp but below the current known
+	    stack segment base, and it was a fault because there was
+	    nothing mapped there (as opposed to a permissions fault),
+	    then extend the stack segment. 
+	 */
+	 Addr base = PGROUNDDN(esp);
+	 Char *ret = VG_(mmap)((Char *)base, seg->addr - base, 
+			       VKI_PROT_READ | VKI_PROT_WRITE | VKI_PROT_EXEC,
+			       VKI_MAP_PRIVATE | VKI_MAP_FIXED | VKI_MAP_ANONYMOUS | VKI_MAP_CLIENT,
+			       -1, 0);
+	 if ((Addr)ret == base) {
+	    VG_(map_segment)(base, seg->addr - base,
+			     VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC,
+			     SF_STACK|SF_GROWDOWN);
+	    return;		/* restart instruction */
+	 }
+	 /* Otherwise fall into normal signal handling */
+      } else if (info->si_code == 2 && /* SEGV_ACCERR */
+		 VG_(needs).shadow_memory &&
+		 VG_(is_shadow_addr)(fault)) {
+	 /* If there's a fault within the shadow memory range, and it
+	    is a permissions fault, then it means that the client is
+	    using some memory which had not previously been used.
+	    This catches those faults, makes the memory accessible,
+	    and calls the tool to initialize that page.
+	 */
+	 static Int recursion = 0;
+
+	 if (recursion++ == 0) {
+	    VG_(init_shadow_range)(PGROUNDDN(fault), VKI_BYTES_PER_PAGE, True);
+	    recursion--;
+	    return;
+	 } else {
+	    /* otherwise fall into normal SEGV handling */	    
+	    recursion--;
+	 }
+      }
+   }
+
    if (VG_(scheduler_jmpbuf_valid)) {
       /* Can't continue; must longjmp back to the scheduler and thus
          enter the sighandler immediately. */
@@ -1661,13 +1730,26 @@ void vg_sync_signalhandler ( Int sigNo, vki_ksiginfo_t *info, struct vki_ucontex
 	 it was actually generated by Valgrind internally.
        */
       struct vki_sigcontext *sc = &uc->uc_mcontext;
+      Char buf[1024];
 
       VG_(message)(Vg_DebugMsg, 
 		   "INTERNAL ERROR: Valgrind received a signal %d (%s) - exiting",
 		   sigNo, signame(sigNo));
+
+      buf[0] = 0;
+      if (1 && !VG_(get_fnname)(sc->eip, buf+2, sizeof(buf)-5)) {
+	 Int len;
+
+	 buf[0] = ' ';
+	 buf[1] = '(';
+	 len = VG_(strlen)(buf);
+	 buf[len] = ')';
+	 buf[len+1] = '\0';
+      }
+
       VG_(message)(Vg_DebugMsg, 
-		   "si_code=%x Fault EIP: %p; Faulting address: %p",
-		   info->si_code, sc->eip, info->_sifields._sigfault._addr);
+		   "si_code=%x Fault EIP: %p%s; Faulting address: %p",
+		   info->si_code, sc->eip, buf, info->_sifields._sigfault._addr);
 
       if (0)
 	 VG_(kill_self)(sigNo);		/* generate a core dump */
