@@ -9,7 +9,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -305,12 +305,20 @@ struct vki_ucontext {
 
 /* The following are copied from include/asm-i386/mman.h .*/
 
+#define VKI_PROT_NONE      0x0		   /* No page permissions */
 #define VKI_PROT_READ      0x1             /* Page can be read.  */
 #define VKI_PROT_WRITE     0x2             /* Page can be written.  */
 #define VKI_PROT_EXEC      0x4             /* Page can be executed.  */
 #define VKI_MAP_ANONYMOUS  0x20            /* Don't use a file.  */
+#define VKI_MAP_SHARED	   0x01		   /* Share changes.  */
 #define VKI_MAP_PRIVATE    0x02            /* Changes are private.  */
 #define VKI_MAP_FIXED      0x10            /* Interpret addr exactly */
+#define VKI_MAP_NOSYMS     0x40000000	   /* internal pseudo-flag to disable symbol loading */
+#define VKI_MAP_CLIENT     0x80000000	   /* internal pseudo-flag to distinguish client mappings */
+
+/* linux/mman.h */
+#define VKI_MREMAP_MAYMOVE	1
+#define VKI_MREMAP_FIXED	2
 
 /* Copied from linux-2.4.19/include/asm-i386/fcntl.h */
 
@@ -366,6 +374,7 @@ struct vki_ucontext {
 #define	VKI_EFAULT          14      /* Bad address */
 #define VKI_EEXIST	    17	    /* File exists */
 #define VKI_EINVAL          22      /* Invalid argument */
+#define VKI_ENFILE          23      /* File table overflow */
 #define VKI_EMFILE	    24	    /* Too many open files */
 #define VKI_ENOSYS          38      /* Function not implemented */
 
@@ -501,6 +510,19 @@ struct vki_pollfd {
 #define VKI_POLLNVAL        0x0020
 
 
+/* sys/epoll.h */
+typedef union vki_epoll_data {
+   void *ptr;
+   Int   fd;
+   UInt  u32;
+   ULong u64;
+} vki_epoll_data_t;
+
+struct vki_epoll_event {
+   UInt events;            /* Epoll events */
+   vki_epoll_data_t data;      /* User data variable */
+};
+
 
 /* 
 ./include/asm-i386/posix_types.h:typedef long           __kernel_suseconds_t;
@@ -571,6 +593,7 @@ struct vki_stat {
    Logic from     /usr/src/linux-2.4.9-31/fs/binfmt_elf.c
                   and its counterpart in the 2.2.14 kernel sources 
                   in Red Hat 6.2.  */
+#define VKI_AT_NULL   0
 #define VKI_AT_SYSINFO 32   /* address of system info page */
 #define VKI_AT_CLKTCK 17    /* frequency at which times() increments */
 #define VKI_AT_HWCAP  16    /* arch dependent hints at CPU capabilities */
@@ -622,12 +645,16 @@ typedef struct vki_modify_ldt_ldt_s {
         unsigned int  limit_in_pages:1;
         unsigned int  seg_not_present:1;
         unsigned int  useable:1;
+        unsigned int  reserved:25;
 } vki_modify_ldt_t;
 
 #define VKI_MODIFY_LDT_CONTENTS_DATA        0
 #define VKI_MODIFY_LDT_CONTENTS_STACK       1
 #define VKI_MODIFY_LDT_CONTENTS_CODE        2
 
+#define VKI_GDT_TLS_ENTRIES 3
+#define VKI_GDT_TLS_MIN     6
+#define VKI_GDT_TLS_MAX     (VKI_GDT_TLS_MIN + VKI_GDT_TLS_ENTRIES)
 
 /* Flags for clone() */
 /* linux/sched.h */
@@ -672,7 +699,17 @@ typedef struct vki_rlimit {
 	unsigned long rlim_max;
 } vki_rlimit;
 
-#define VKI_RLIMIT_NOFILE 7
+#define VKI_RLIMIT_CPU		0		/* CPU time in ms */
+#define VKI_RLIMIT_FSIZE	1		/* Maximum filesize */
+#define VKI_RLIMIT_DATA		2		/* max data size */
+#define VKI_RLIMIT_STACK	3		/* max stack size */
+#define VKI_RLIMIT_CORE		4		/* max core file size */
+#define VKI_RLIMIT_RSS		5		/* max resident set size */
+#define VKI_RLIMIT_NPROC	6		/* max number of processes */
+#define VKI_RLIMIT_NOFILE	7		/* max number of open files */
+#define VKI_RLIMIT_MEMLOCK	8		/* max locked-in-memory address space */
+#define VKI_RLIMIT_AS		9		/* address space limit */
+#define VKI_RLIMIT_LOCKS	10		/* maximum file locks held */
 
 /* Socket stuff. */
 /*
@@ -717,6 +754,15 @@ struct statfs64 {
         unsigned int f_spare[5];
 };
 
+/* 
+ * linux/futex.h
+ */
+
+#define VKI_FUTEX_WAIT    0
+#define VKI_FUTEX_WAKE    1
+#define VKI_FUTEX_FD      2
+#define VKI_FUTEX_REQUEUE 3
+
 
 #endif /*  __VG_KERNELIFACE_H */
 
